def determine_rows_cols_for_subplots(num_subplots):
  # Determine the optimal number of rows and columns for a 'plt.subplots'
  # call based on the number of subplots in the figure, preferring a 
  # square layout. The routine favours the number closest to the 
  # number of subplots, with a minimum of 'num_subplots'
  from itertools import product
  import numpy as np

  # Determine min and max rows to optimise matrix size
  root = np.sqrt(num_subplots)

  # Min and max size for the matrix indices
  max_size = int(np.ceil(root) + 1)
  min_size = int(np.floor(root) - 1)
  matrix_size = max_size - min_size

  matrix = np.zeros(shape=(matrix_size, matrix_size))
  r = range(min_size, max_size)

  for ((indx, x), (indy, y)) in product(enumerate(r), enumerate(r)):
    matrix[indx, indy] = x * y

  # Set elements less than 'num_subplots' to infinity to filter them out
  matrix[np.where(matrix < num_subplots)] = np.inf
  nrows, ncols = np.unravel_index(np.argmin(matrix), matrix.shape)

  # Add 'min_size' back to indices to scale it to normal 'matrix'
  nrows += min_size
  ncols += min_size

  return nrows, ncols



import matplotlib.pyplot as plt
import numpy as np
import .plotHelper as pH


try:
  plt.style.use('scientific_grid_no_space')
except:
  pass


# -------------------------------------------------------------------------
# Modular plotting functions
# -------------------------------------------------------------------------
def plot_data(ax, x_data, y_data, x_error_data=None, y_error_data=None,
    color='k', ls='-',
    xlabel=None, ylabel=None,
    xscale='linear', yscale='linear'):
  # Plot 'x_' and 'y_data' on specified axis 'ax', including errors if
  # present
  fig = None
  if ax == None:
    fig, ax = plt.subplots()

  # Plot data
  if x_error_data or y_error_data:
    ax.errorbar(x_data, y_data, xerr=x_error_data, yerr=y_error_data,
       color=color, ls=ls)
  else:
    ax.plot(x_data, y_data, color=color, ls=ls)

  # Labels
  if xlabel:
    ax.set_xlabel(xlabel)
  
  if ylabel:
    ax.set_ylabel(ylabel)

  # Set scales
  ax.set_xscale(xscale)
  ax.set_yscale(yscale)

  return fig, ax  # 'fig' will be None if 'ax' was provided

def plot_histogram(ax, data, color='gray', bins=None,
                       xlabel=None, ylabel=None, xscale="linear", yscale="linear"):
  # Plot a histogram of the specified DataFrame column 'df[key]'
  fig = None
  if ax == None:
    fig, ax = plt.subplots()

  mean, median, std_dev = np.mean(data), np.nanmedian(data), np.std(data)

  ax.hist(data, bins=bins, density=True, color=color)

  # Vertical lines for mean, median
  ax.axvline(mean, c='r', ls='--')
  ax.axvline(median, c='b', ls='--')

  # Set scales
  ax.set_xscale(xscale)
  ax.set_yscale(yscale)

  if xlabel:
    ax.set_xlabel(xlabel)

  if ylabel:
    ax.set_ylabel(ylabel)

  return fig, ax  # 'fig' will be None if 'ax' was provided


def plot_data_correlations(ax, correlations_array, 
    cmap='jet', add_text=True, xlabels=[], ylabels=[]):
  # Plot correlations between parameters as a heatmap
  fig = None
  if ax == None:
    fig, ax = plt.subplots()

  # Plot heatmap
  ax.imshow(correlations_array, cmap=cmap)
  
  # Axes ticks
  ticks = np.arange(len(correlations_array))
  ax.set_xticks(ticks)
  ax.set_yticks(ticks)
  
  # Axes tick labels
  ax.set_xticklabels(xlabels)
  ax.set_yticklabels(ylabels)

  # Rotate the tick labels and set their alignment.
  plt.setp(ax.get_xticklabels(), rotation=45, ha="right",
          rotation_mode="anchor")

  # Add text annotations of correlation coefficients to heatmap
  if add_text:
    for i in range(len(correlations_array)):
      for j in range(len(correlations_array)):
        ax.text(j, i, f"{correlations_array[i, j]:.3f}", ha="center", va="center", color="w")

  return fig, ax  # 'fig' will be None if 'ax' was provided


# -------------------------------------------------------------------------
# Dataframe plotting functions
# -------------------------------------------------------------------------
def plot_df_key(ax, df, x_key, y_key, x_error_key=None, y_error_key=None, 
    color='k', ls='-',
    xlabel=None, ylabel=None, 
    xscale="linear", yscale="linear"):
  # Plot DataFrame columns 'df[x_key]' vs 'df[y_key]'
  fig = None
  if ax == None:
    fig, ax = plt.subplots()

  # Copy DF for plotting so it is not changed outside of this scope
  plot_df = df.copy()

  # Get data from DF
  x_data = plot_df[x_key]
  y_data = plot_df[y_key]

  x_error_data = None
  y_error_data = None

  # Get error data from DF
  if x_error_key:
    x_error_data = plot_df[x_error_key]
  
  if y_error_key:
    y_error_data = plot_df[y_error_key]

  # Plot data
  fig, ax = plot_data(ax, x_data, y_data, x_error_data=x_error_data, y_error_data=y_error_data,
    color=color, ls=ls, xlabel=xlabel, ylabel=ylabel, xscale=xscale, yscale=yscale)

  return fig, ax  # 'fig' will be None if 'ax' was provided


def plot_df_key_histogram(ax, df, key, color='gray', bins=None,
                       xlabel=None, ylabel=None, xscale="linear", yscale="linear"):
  # Plot a histogram of the specified DataFrame column 'df[key]'
  fig = None
  if ax == None:
    fig, ax = plt.subplots()

  # Copy DF for plotting so it is not changed outside of this scope
  plot_df = df.copy()

  data = plot_df[key]

  fig, ax = plot_histogram(ax, data, color=color, bins=bins, 
    xlabel=xlabel, ylabel=ylabel, xscale=xscale, yscale=yscale)

  return fig, ax  # 'fig' will be None if 'ax' was provided


def plot_df_correlations(ax, df, keys, cmap='jet', add_text=True):
  # Calculate and plot correlations between 'keys' in DataFrame 'df'
  import utilities.dataframeUtilities as dfut

  fig = None
  if ax == None:
    fig, ax = plt.subplots()

  plot_df = df.copy()
  correlations = dfut.get_correlation(plot_df, columns=keys)

  fig, ax = plot_data_correlations(ax, correlations, 
    cmap=cmap, add_text=add_text, xlabels=keys, ylabels=keys)

  return fig, ax  # 'fig' will be None if 'ax' was provided



import operator as opt
import functools as ft

# -------------------------------------------------------------------------
# List utility functions
# -------------------------------------------------------------------------
def list_instances_to_dict(list_):
  # Takes in 'reactants' or 'products' list which may have duplicate keys
  # and creates the dictionary expected by Reaction()
  output_dict = {}
  for element in [sublist for i, sublist in enumerate(list_, 1) if sublist not in list_[i:]]:
    output_dict[element] = list_.count(element)

  return output_dict


def reduce_list(list_, operator=opt.mul):
  # Reduce a list to a single value based on the operator provided
  return ft.reduce(lambda a, b: operator(a, b), list_)


def chunk_list(list, n):
  # Subdivide and return list in chunks based on sample 'n'
  chunks = []
  step = len(list) // n - 1

  for i in range(0, len(list), step):
      chunks.append(list[i: i + step])

  return chunks



# -------------------------------------------------------------------------
# Dictionary utility functions
# -------------------------------------------------------------------------
def convert_log_dict_to_linear(log_dict):
  linear_dict = {}

  for key, value in log_dict.items():
    linear_dict[key] = 10**value
  
  return linear_dict


def add_unit_to_dict_items(dictionary, unit):
  unit_dict = {}

  for key, value in dictionary.items():
    unit_dict[key] = value * unit

  return unit_dict



import pandas as pd


def get_df(filepath, filetype='infer', index=None):
  if filetype == 'infer':
    filetype = filepath.split('.')[-1]

  # Read in file as DataFrame
  if filetype == 'csv':
    df = pd.read_csv(filepath)
  elif filetype == 'pickle' or filetype == 'pkl':
    df = pd.read_pickle(filepath)
  else:
    print(f"Error, filetype {filetype} not understood.")
    return -1

  # Set index
  if index:
    df = df.set_index(index)
    df.sort_index(ascending=True, inplace=True)

  return df

def concat_dfs(df1, df2, how='inner'):
  # Assumes DataFrames have the same index
  concatenated_df = pd.merge(df1, df2, how=how)

  return concatenated_df

def truncate_year_range(df, year_range):
  # Assumes 'df' indexed by Date
  return df[df['DecimalDate'].between(year_range[0], year_range[1])].copy()
  
def get_correlation(df, columns):
    # Pearson correlation coefficient
    corr_df = df[columns].copy()
    
    return corr_df.corr(method='pearson')



import numpy as np


# -------------------------------------------------------------------------
# Functions for spectroscopy
# -------------------------------------------------------------------------
def fwhm_to_std_dev(fwhm):
    return fwhm / (np.sqrt(8 * np.log(2)))


def std_dev_to_fwhm(std_dev):
    return np.sqrt(8 * np.log(2)) * std_dev


def velocities_to_wavelengths(velocities, line_centre):
    wavelengths = line_centre * ((np.array(velocities) / 2.99792458e5) + 1.)
    return wavelengths


def wavelengths_to_velocities(wavelengths, line_centre):
    velocities = ((np.array(wavelengths) / line_centre) - 1.) * 2.99792458e5
    return velocities


import numpy as np
import utilities as util
import astropy.units as u


# -------------------------------------------------------------------------
# Derived Units
# -------------------------------------------------------------------------
cm3 = u.cm**3
dm3 = u.dm**3
m3 = u.m**3
g_cm3 = (u.g / cm3)
molar = u.mol / dm3

# -------------------------------------------------------------------------
# Constants
# -------------------------------------------------------------------------
avogadro_constant = 6.0221415E+23 / u.mol  # atoms [mol^-1]
hydrogen_mass_amu = 1.008  # a.m.u., standard atomic weight
hydrogen_mass_cgs = 1.6733E-24 * u.g  # [g]

# -------------------------------------------------------------------------
# Conversion functions for physical quantities
# -------------------------------------------------------------------------
def gas_density_to_hydrogen_number_density(gas_density, percentage_hydrogen=1, log=False):
  # Converts gas density (assumed [g cm^-3]) to a hydrogen number density 
  # based on the percentage of hydrogen (default 100% hydrogen)
  hydrogen_number_density = (gas_density / hydrogen_mass_cgs) * percentage_hydrogen

  if log:
    return np.log10(hydrogen_number_density.value)
  else:
    return hydrogen_number_density

def log_abundance_to_number_density(log_abundance, log_hydrogen_number_density, log_hydrogen_abundance=12):
  # Using the log(hydrogen abundance) (default '12' for solar physics),
  # convert from stellar photospheric abundance to a number density
  log_number_density = log_abundance - log_hydrogen_abundance + log_hydrogen_number_density
  number_density = 10**log_number_density
  
  return number_density

def number_density_to_concentration(number_density):
  # Assumes number density in [cm^-n] & 'avogadro_constant' in [mol^-1]
  # so concentration ~ [mol cm^-n]
  concentration = number_density / avogadro_constant

  return concentration

def concentration_to_number_density(concentration):
  # Assumes concentration ~ [mol cm^-3] & 'avogadro_constant' in [mol^-1]
  # so number density in [cm^-3]
  number_density = concentration * avogadro_constant

  return number_density
  
# -------------------------------------------------------------------------
# Stoichiometry functions
# -------------------------------------------------------------------------
def calculate_stoichiometry(reactants_list, products_list, return_dicts=False):
  reactant_stoichiometry = util.list_instances_to_dict(reactants_list)
  product_stoichiometry = util.list_instances_to_dict(products_list)

  if return_dicts:
    return reactant_stoichiometry, product_stoichiometry
  else:
    return reactant_stoichiometry.values, product_stoichiometry.values

# -------------------------------------------------------------------------
# Reaction rate and timescale functions
# -------------------------------------------------------------------------
def modified_to_arrhenius_prefactor(alpha, beta, temperature):
	# Modified Arrhenius rate prefactor to Arrhenius rate prefactor
	return alpha * (temperature / 300)**beta

def modified_arrhenius_rate(alpha=1, beta=0, gamma=0, temperature=300):
  # rate = alpha * (T/300[K])^beta * EXP(-gamma / T)
	# Equivalent to standard Arrhenius for beta=0
  return alpha * (temperature/300)**beta * np.exp(-gamma / temperature)

def arrhenius_rate(prefactor, activation_energy, temperature):
  return prefactor * np.exp(-activation_energy / temperature)

def calculate_unitless_timescale(forward_rate, reverse_rate=None, reactant_stoichiometry=[1, 1]):
  # Example case: A + B -> C
  # Note that this is only valid if rates are dimensionless. Othwerwise, 
  # the rates need to be related to one another based on the number 
  # densities of the reactants and products
  total_rate = forward_rate.value * np.sum(reactant_stoichiometry)

  if reverse_rate != None:
    total_rate += reverse_rate.value
  
  timescale = 1 / total_rate

  return timescale

# -------------------------------------------------------------------------
# Functions for determining equilibrium quantities
# -------------------------------------------------------------------------
def equilibrium_constant_dimensional(forward_rate_coefficient,
                                    reverse_rate_coefficient):
  # dimensional equilibrium constant K_eq = k_f / k_r
  equilibrium_constant = forward_rate_coefficient / reverse_rate_coefficient

  return equilibrium_constant

# -------------------------------------------------------------------------
# Utility functions for rates and timescales
# -------------------------------------------------------------------------
def scale_timescale_with_number_density(timescale, reactant_number_density):
  # Scale timescale based on number of primary reactants; Wedemeyer-Boehm(2005)
  scaled_timescale = timescale / reactant_number_density
  
  return scaled_timescale

def convert_rate_to_molar(rate, order):
  # Assumes rate is in [cm^-3*n s^-1] where 'n' is order of reaction
  # Convert from cm^3 -> dm^3 & use Avogadro number to change from 
  # volume / atom -> volume / mol
  volume_conversion = cm3.to(dm3)
  conversion_factor = (volume_conversion * avogadro_constant)**order
  molar_rate = rate * conversion_factor  # units of [mol^n dm^-3n s^-1]

  return molar_rate

# -------------------------------------------------------------------------
# Unit conversion functions
# -------------------------------------------------------------------------
def get_rate_unit_from_order(order, unit_system='cgs'):
  # Based on the order of the reaction (n), determine the proper units for 
  # the reaction rate.
  unit = None
  if unit_system == 'cgs':
    # [cm^3*n s^-1]
    unit = (cm3)**order
  elif unit_system == 'molar':
    # [M^(1-n) s^-1]
    unit = (u.M)**(1-order)
  
  unit /= u.s

  return unit


import setuptools

with open('README.md', 'r') as fh:
  long_description = fh.read()

setuptools.setup(
  name="sd-data-tools",
  version="0.0.1",
  author="Siddhant A. Deshmukh",
  author_email="siddhant593@gmail.com",
  description="A collection of useful plotting, analysis and manipulation scripts for data science.",
  long_description=long_description,
  url="https://github.com/SiddhantDeshmukh/usefulScripts",
  packages=setuptools.find_packages(),
  classifiers=[
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
  ],
  python_requires=">=3.6"
)



import numpy as np
from scipy.interpolate import splrep, splev
import math


# -------------------------------------------------------------------------
# Grid functions
# -------------------------------------------------------------------------
def create_grid(x_range, y_range, num_x_points=50, num_y_points=50):
  # Utility function for creating a general mesh grid
  x = np.linspace(*x_range, num=num_x_points)
  y = np.linspace(*y_range, num=num_y_points)

  grid = np.meshgrid(x, y)

  return grid


# -------------------------------------------------------------------------
# Interpolation functions
# -------------------------------------------------------------------------
def determine_fits(x_list, data_list):
  # Takes in a single x-variable list and a list of y-variable lists to generate fits for each
    fits = []
    # x_list = sorted(x_list)

    for data in data_list:
        fit = splrep(x_list, data)
        fits.append(fit)

    return fits


def evaluate_fits(x_list, fits_list):
  # Takes in a list of x-value lists and an associated fits list for interpolation
    y_data_list = [[] for _ in range(len(fits_list))]

    for x in x_list:
        for index, fit in enumerate(fits_list):
            y_data_list[index].append(float(splev(x, fit)))

    return y_data_list


# -------------------------------------------------------------------------
# Arithmetic utility functions
# -------------------------------------------------------------------------
def even_to_odd(number, direction='up'):
    """
    Casts an even number to an odd number. Used mainly to centre axis labels for plots by having an odd number of
    rows and columns.
    
    Arguments:
        number {Int} -- Even number to be cast to an odd number.
    
    Keyword Arguments:
        direction {str} -- If 'up', add 1 to the even number. If 'down', subtract 1 from the even number. 
        (default: {'up'})
    
    Returns:
        {int} -- Odd number
    """
    number = int(number)
    if number % 2 == 0:
        if direction is 'up':
            number += 1
        elif direction is 'down':
            number -= 1
        else:
            print("Invalid option {direction}. Valid options are \'up\' or \'down\'. Choosing \'up\'.")
            number += 1
    
    return number


def is_square(number: int) -> bool:
  root = math.isqrt(number)
  return number == root ** 2
